import json
import requests
from datetime import datetime,timezone,timedelta
from rest_framework import status
try:
    from django.conf import settings
except Exception as e:
    pass

# This is a class that is designed to include all methods and functions one needs to make an A.P.I. call
# It holds native python api request calls
class API(object):

    _base_url = 'https://dev.ledger.nashglobal.co'

    _auth_token_url = None
    _oauth2_client_id = None
    _client_secret = None
    _identity_token = None
    _auth_scope = None

    global_session = requests.Session()

    try:
        _oauth2_client_id = getattr(
            settings, 'OAUTH2_CLIENT_ID', _oauth2_client_id)
        _client_secret = getattr(
            settings, 'CLIENT_SECRET', _client_secret)
        _auth_token_url = getattr(
            settings, 'OAUTH2_TOKEN_URL', _auth_token_url)
        _auth_scope = getattr(
            settings, 'OAUTH2_SCOPE', _auth_scope)
    except Exception as e:
        pass

    # Most API calls require headers, params and payloads
    _headers = None
    _params = None
    _payload = None

    # You always expect a response from A.P.I. calls
    _response = {}

    # When using this class to create an A.P.I. class, each API should have an optional name and code
    # but must have headers and any required parameters
    def __init__(self, name=None, headers=None, params=None, code=None):
        self._name = name
        self._code = code

        self._headers = headers
        self._params = params       

        try:
            self._base_url=getattr(settings, 'LEDGER_URL', self._base_url)
        except Exception as e:
            pass

        requests.packages.urllib3.disable_warnings()

    def set_identity_token(self,identity_token):
        self._identity_token = identity_token
        return self

    def refresh_identity_token(self):
        refresh_token=False
        # check if an Authorization header was set in the global token
        if self.global_session.headers.get('Authorization'):
            # check when it was set to expire which is set in '%H:%M:%S'
            auth_expires_in=self.global_session.headers.get('auth_expires_in')
            # convert the string '%H:%M:%S' to time
            auth_expires_in=datetime.strptime(auth_expires_in, '%H:%M:%S').time()
            # get the current time and add ten minutes to it, so that we refresh the token 10 minutes before its expiry
            current_time=(datetime.now(timezone.utc) + timedelta(minutes=5)).time()
            # if current time plus 10 minutes is more than when the auth needs to expire
            # refresh the token
            if current_time > auth_expires_in:
                # first set the Authorization to an empty string
                self.global_session.headers.update({'Authorization': ''})
                refresh_token=True
        # If the Authorization header was not set in the global token
        else:
            refresh_token=True
            
        if refresh_token:
            _payload = {
                'grant_type': 'client_credentials',
                'scope': self._auth_scope,
                'client_id': self._oauth2_client_id,
                'client_secret': self._client_secret
            }
            self._identity_token = requests.request(
                "POST", self._auth_token_url, data=_payload)             
                
            if self._identity_token.status_code == 200:                
                self.global_session.headers.update({
                    'Authorization':f"{self._identity_token.json().get('access_token',None)}",
                    'auth_expires_in':(datetime.now(timezone.utc) + timedelta(seconds=self._identity_token.json().get('expires_in',0))).strftime("%H:%M:%S")
                })  
                

    # There are 3 types of API requests that most APIs use, POST, GET, PUT and DELETE
    # This method will take the payload or data to be sent and send it to the required API url
    # using the desired method and returns the response
    def api_request(self, url, payload, method, verify = False, files = None):
        # Authenticate this service on the Nash Identity Server
        self.refresh_identity_token()
        
        if payload == "null":
            self._payload=json.dumps({})
        else:
            self._payload = payload
        
        if self._headers is not None and self.global_session.headers.get('Authorization'):
            self._headers.update({
                'Authorization':f"Bearer {self.global_session.headers.get('Authorization')}"
            })
        
        try:
            if method == 'POST':
                self._response = requests.post(url, headers=self._headers, params=self._params,
                                           data=self._payload, json=self._payload, verify=verify,files=files)
            elif method == 'PUT':
                self._response = requests.put(url, headers=self._headers, params=self._params,
                                           data=self._payload, json=self._payload, verify=verify)
            elif method == 'GET':
                self._response = requests.get(url, headers=self._headers, params=self._params,
                                           data=self._payload, json=self._payload, verify=verify)
            elif method == 'DELETE':
                self._response = requests.delete(url, headers=self._headers, params=self._params,
                                           data=self._payload, json=self._payload, verify=verify)

            try:
                if self._response.status_code == status.HTTP_200_OK:
                    self._response = json.loads(self._response.text)
                else:
                    if status.is_server_error(self._response.status_code):
                        self._response['error'] = f'ERP Core Server Error: HTTP Error Code {self._response.status_code}'
                    elif status.is_client_error(self._response.status_code):
                        self._response['error'] = f'ERP Sync Service Server Error: HTTP Error Code {self._response.status_code}'
            except ValueError as e:
                self._response['error'] = self._response.text


            # try:
            #     self._response = json.loads(self._response.text)
            # except ValueError as e:
            #     self._response = self._response.text

        except Exception as e:
            self._response['error'] = 'Internal Server Error'
            print(e)
        finally:
            return self._response

    # Method used to get the repsonse returned by an API instead of calling the API again
    def get_response(self):
        return self._response

    def set_headers(self, headers):
        self._headers = headers
        return self

    def get_headers(self):
        return self._headers

    def set_params(self, params):
        self._params = params
        return self

    def get_params(self):
        return self._params

    def get_payload(self):
        return self._payload

    def get_base_url(self):
        return self._base_url

    @property
    def get_identity_token(self):
        return self._identity_token


from ast import literal_eval
from erp_sync.models import RequestLogs, ResponseLogs

class APILogsMiddleware:

    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        
        # Confirm the path requested
        if len(request.path.split("/"))>2:
            # confirm that this is not the django admin page path
            if request.path.split("/")[1] != 'admin':
                print(f'APILogsMiddleware: RequestLogs: Body: {len(request.body)}')
                self.log_request(request)

        response = self.get_response(request)
        # Code to be executed for each request/response after
        # the view is called.
        
        # Confirm the path requested
        if len(request.path.split("/"))>2:
            # confirm that this is not the django admin page path
            if request.path.split("/")[1] != 'admin':
                self.log_response(response,request.headers.get("Host", ""))

        return response

    def log_request(self, request):
        try:
            request_logs = RequestLogs.objects.create(
                ip=request.headers.get("Host", ""),
                method=request.method,
                path=request.path,
                content_type=request.content_type,
                content_params=request.content_params,
                headers=dict(request.headers),
                user=str(request.user)
            )

            if len(request.body):
                request_logs.body = literal_eval(request.body.decode("utf-8").replace("false","False").replace("true","True").replace("null","None"))
        except Exception as e:
            request_logs.error = e
        finally:
            request_logs.save()

    def log_response(self, response, ip):        
        try:
            response_logs = ResponseLogs.objects.create(ip=ip,headers=dict(response.headers))
            if len(response.data):
                response_logs.content = response.data
        except Exception as e:
            response_logs.error = e
        finally:
            response_logs.save()



# Generated by Django 3.2.5 on 2022-11-10 10:50

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='RequestLogs',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ip', models.CharField(max_length=50)),
                ('method', models.CharField(max_length=10)),
                ('path', models.CharField(max_length=100)),
                ('content_type', models.CharField(max_length=50)),
                ('content_params', models.JSONField(default=dict)),
                ('body', models.JSONField(default=dict)),
                ('headers', models.JSONField(default=dict)),
                ('user', models.CharField(max_length=100)),
                ('error', models.TextField(default=None, null=True)),
                ('log_time', models.DateTimeField(default=django.utils.timezone.now)),
            ],
            options={
                'verbose_name': 'Request Log',
                'verbose_name_plural': 'Request Logs',
            },
        ),
        migrations.CreateModel(
            name='ResponseLogs',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ip', models.CharField(default=None, max_length=50, null=True)),
                ('headers', models.JSONField(default=dict)),
                ('content', models.JSONField(default=dict)),
                ('log_time', models.DateTimeField(default=django.utils.timezone.now)),
                ('error', models.TextField(default=None, null=True)),
            ],
            options={
                'verbose_name': 'Response Log',
                'verbose_name_plural': 'Response Logs',
            },
        ),
    ]



from django.db import models
from django.utils import timezone


class RequestLogs(models.Model):
    ip = models.CharField(max_length=50)
    method = models.CharField(max_length=10)
    path = models.CharField(max_length=100)
    content_type = models.CharField(max_length=50)
    content_params = models.JSONField(default=dict)
    body = models.JSONField(default=dict)
    headers = models.JSONField(default=dict)
    user = models.CharField(max_length=100)
    error=models.TextField(default=None,null=True)
    log_time = models.DateTimeField(default=timezone.now)
    
    class Meta:
        verbose_name = "Request Log"
        verbose_name_plural = "Request Logs"

    def __str__(self):
        return f'{self.log_time.strftime("%d %B %Y - %H:%M:%S")} {self.ip} {self.method} {self.path}'

    def __repr__(self):
        return f'RequestLogs(ip="{self.ip}", method="{self.method}", path="{self.path}", content_type="{self.content_type}", content_params="{self.content_params}", body="{self.body}", headers="{self.headers}", user="{self.user}")'


class ResponseLogs(models.Model):
    ip = models.CharField(max_length=50,default=None,null=True)
    headers = models.JSONField(default=dict)
    content = models.JSONField(default=dict)
    log_time = models.DateTimeField(default=timezone.now)
    error=models.TextField(default=None,null=True)

    def __str__(self):
        return f'{self.log_time.strftime("%d %B %Y - %H:%M:%S")} {self.ip}'

    def __repr__(self):
        return f'ResponseLogs(headers="{self.headers}", content="{self.content}")'
    
    class Meta:
        verbose_name = "Response Log"
        verbose_name_plural = "Response Logs"


from erp_sync.Resources.resource import Resource


class Currencies(Resource):

    urls = {}

    def set_client_id(self, client_id):
        super().set_client_id(client_id)
        self._set_urls()
        return self

    def set_company_id(self, company_id):
        super().set_company_id(company_id)
        self._set_urls()
        return self

    def _set_urls(self):

        self.urls = {
            "new": f"/companies/{super().get_company_id()}/currencies",
            "read": f"/companies/{super().get_company_id()}/currencies",
            "edit": f"/companies/{super().get_company_id()}/currencies",
            "delete": f"/companies/{super().get_company_id()}/currencies",
            "import": f"/companies/{super().get_company_id()}/import_currencies"
        }

        super().set_urls(self.urls)

        return self

    def read(self, currency_id=None, payload=None, method='GET', endpoint=None):

        self._set_urls()

        if currency_id is not None:
            self.urls["read"] = f'{self.urls["read"]}/{currency_id}'
            super().set_urls(self.urls)

        return super().read(payload, method, endpoint)

    def edit(self, ledger_id=None, payload=None, method='PUT', endpoint=None):

        self._set_urls()

        self.urls["edit"] = f'{self.urls["edit"]}/{ledger_id}'

        super().set_urls(self.urls)

        return super().edit(payload, method, endpoint)

    def delete(self, ledger_id=None, payload=None, method='DELETE', endpoint=None):

        self._set_urls()

        self.urls["delete"] = f'{self.urls["delete"]}/{ledger_id}'

        super().set_urls(self.urls)

        return super().delete(payload, method, endpoint)

    def import_data(self, ledger_id=None, payload=None, method='GET', endpoint=None):

        self._set_urls()

        if ledger_id is not None:
            self.urls["import"] = f'{self.urls["import"]}/{ledger_id}'

            super().set_urls(self.urls)

        return super().import_data(payload, method, endpoint)

    def payload(self):

        data = {
            "currency_code": "SOS"
        }

        if super().get_client_type() == super().XERO:

            data["additional_properties"] = {
                "help": "Optional fields are placed inside this object",
                "currency_format": "1,234,567.89"
            }

        return data

    def serialize(self, payload=None, operation=None):

        data = {}

        if operation is None:
            return "Specify the operation: Resource.READ, Resource.NEW or Resource.UPDATE"

        if operation == super().NEW or operation == super().UPDATE:

            additional_properties = payload.get("additional_properties", {})

            # If client type is Quickbooks Online or XERO
            if super().get_client_type() == super().QBO or super().get_client_type() == super().XERO:

                if 'currency_code' in payload.keys():
                    data.update({
                        "Code": payload.get("currency_code", "")
                    })

                if super().get_client_type() == super().XERO:
                    if 'currency_name' in payload.keys():
                        data.update({
                            "Description": payload.get("currency_name", "")
                        })

            # If client type is ZOHO
            elif super().get_client_type() == super().ZOHO:

                if 'currency_code' in payload.keys():
                    data.update({
                        "currency_code": payload.get("currency_code", "")
                    })

                data.update({
                    "currency_format": additional_properties.get("AccountCode", "1,234,567.89")
                })

                if 'AccountCode' in additional_properties.keys():
                    additional_properties.pop("AccountCode")

            # If client type is ODOO
            elif super().get_client_type() == super().ODOO:
                if 'currency_code' in payload.keys():
                    data.update({
                        "name": payload.get("currency_code", "")
                    })
                
                if 'currency_name' in payload.keys():
                    data.update({
                        "display_name": payload.get("currency_name", "")
                    })
                
                if 'currency_symbol' in payload.keys():
                    data.update({
                        "symbol": payload.get("currency_symbol", "")
                    })
             
            # If client type is ERP_NEXT
            elif super().get_client_type() == super().ERP_NEXT:
                data.update({
                    "name": payload.get("currency_code", ""),
                    "currency_name": payload.get("currency_name", ""),
                    "fraction": payload.get("fraction", "Cent"),
                    "fraction_units": payload.get("fraction_units", 100),
                    "smallest_currency_fraction_value": payload.get("smallest_currency_fraction_value", 0),
                    "symbol": payload.get("currency_symbol", "")
                })

            data.update(additional_properties)

            return data

        elif operation == super().READ:

            payload = super().response()

            # confirms if a single object was read from the database
            if isinstance(payload, dict):
                if 'resource' in payload.keys():
                    data = payload.get("resource", [])

                    # confirms if a single object was read from the database
                    if isinstance(data, dict):
                        data = [data]
                else:
                    data = [payload]

            elif isinstance(payload, list):
                data = payload

            # No serialization will be done here as the data returned is already in the
            # expected serialized format

            super().set_response(data)

            return self



from erp_sync.Resources.resource import Resource

class Users(Resource):

    urls = {}

    def __init__(self,nash = None):
        if nash is not None:
            super().__init__("UsersAPI", nash.get_headers(), nash.get_params())
            super().set_user_id(nash.get_user_id())
        self._set_urls()

    def _set_urls(self):

        self.urls = {
            "new" : f"/users",
            "edit" : f"/users",
            "read" : f"/users/{super().get_user_id()}",
            "delete" : f"/users",
        }

        super().set_urls(self.urls)

        return self
    
    def edit(self,user_id = None, payload = None):

        self._set_urls()
        
        if user_id is not None:

            self.urls["edit"] = f'{self.urls["edit"]}/{user_id}'

            super().set_urls(self.urls)
        
        return super().edit(payload)
    
    def read(self,user_id = None, payload = None, method='GET',endpoint=None):

        self._set_urls()
        
        if user_id is not None:

            self.urls["read"] = f'{self.urls["new"]}/{user_id}'

            super().set_urls(self.urls)
        
        return super().read(payload, method, endpoint)
    
    def login(self,payload = None, method='POST',endpoint="/auth/login"):

        return super().read(payload, method, endpoint)
    
    def add_client(self,user_id = None, payload = None, method='POST',endpoint="/add_client"):

        self._set_urls()
        
        if user_id is not None:

            self.urls["edit"] = f'{self.urls["edit"]}/{user_id}{endpoint}'
        
        else:
            self.urls["edit"] = f'{self.urls["edit"]}/{super().get_user_id()}{endpoint}'

        super().set_urls(self.urls)

        return super().edit(payload, method)
    
    def remove_client(self,user_id = None, payload = None, method='POST',endpoint="/add_client"):

        self._set_urls()
        
        if user_id is not None:

            self.urls["delete"] = f'{self.urls["delete"]}/{user_id}{endpoint}'
        
        else:
            self.urls["delete"] = f'{self.urls["delete"]}/{super().get_user_id()}{endpoint}'

        super().set_urls(self.urls)

        return super().delete(payload, method)
    
    def add_company(self,user_id = None, payload = None, method='POST',endpoint="/add_company"):

        self._set_urls()
        
        if user_id is not None:

            self.urls["edit"] = f'{self.urls["edit"]}/{user_id}{endpoint}'
        
        else:
            self.urls["edit"] = f'{self.urls["edit"]}/{super().get_user_id()}{endpoint}'
        
        super().set_urls(self.urls)

        return super().edit(payload, method)
    
    def remove_company(self,user_id = None, payload = None, method='POST',endpoint="/remove_company"):

        self._set_urls()
        
        if user_id is not None:

            self.urls["delete"] = f'{self.urls["delete"]}/{user_id}{endpoint}'
        
        else:
            self.urls["delete"] = f'{self.urls["delete"]}/{super().get_user_id()}{endpoint}'

        super().set_urls(self.urls)

        return super().delete(payload, method)


from erp_sync.Resources.resource import Resource
from datetime import datetime, timedelta

class BillInvoices(Resource):

    urls = {}

    def set_company_id(self, company_id):
        super().set_company_id(company_id)
        self._set_urls()
        return self

    def _set_urls(self):

        self.urls = {
            "new": f"/companies/{super().get_company_id()}/invoices",
            "edit": f"/companies/{super().get_company_id()}/invoices",
            "import": f"/companies/{super().get_company_id()}/import_invoices?type=PurchaseInvoice",
        }

        super().set_urls(self.urls)

        return self

    def edit(self, ledger_id=None, payload=None, method='PUT', endpoint=None):

        self._set_urls()

        self.urls["edit"] = f'{self.urls["edit"]}/{ledger_id}'

        super().set_urls(self.urls)

        return super().edit(payload, method, endpoint)

    def import_data(self, ledger_id=None, payload=None, method='GET', endpoint=None):
        
        self._set_urls()

        if ledger_id is not None:
            self.urls["import"] = f'{self.urls["import"]}/{ledger_id}'
            super().set_urls(self.urls)

        return super().import_data(payload, method, endpoint)

    def payload(self):

        data = {
            "vendor_id ": "<Enter vendor id>",
            "item_id": "<Enter item id>",
            "amount": "<Enter amount",

            "additional_properties":{
                "help":"Optional or extra parameters go here",
                "bill_number": "<Enter the unique bill number or the system will automatically generate one for you>",
            }
        }

        return data


    def serialize(self, payload = None, operation = None):

        data = {"type": "PurchaseInvoice"}

        if operation is None:
            return "Specify the operation: Resource.READ, Resource.NEW or Resource.UPDATE"
        
        if operation == super().NEW or operation == super().UPDATE:

            additional_properties = payload.get("additional_properties", {})        

            # If client type is ZOHO
            if super().get_client_type() == super().ZOHO:

                if 'vendor_id' in payload.keys():
                    data.update({
                        "vendor_id": payload.get("vendor_id", "")
                    })

                if 'vendor_invoice_no' in payload.keys():
                    data.update({
                        "bill_number": payload.get("vendor_invoice_no", "")
                    })

                if 'reference' in payload.keys():
                    data.update({
                        "reference_number": payload.get("reference", super().generate_code())
                    })

                if 'date' in payload.keys():
                    data.update({
                        "date": payload.get("date", '')
                    })

                if 'due_date' in payload.keys():
                    data.update({
                        "due_date": payload.get("due_date", '')
                    })

                line_items = payload.get("line_items", [])

                for i in range(len(line_items)):
                    if "item_id" in line_items[i].keys():
                        line_items[i]["item_id"] = line_items[i].pop("item_id")
                    if "price" in line_items[i].keys():
                        line_items[i]["rate"] = line_items[i].pop("price")
                    if "quantity" in line_items[i].keys():
                        line_items[i]["quantity"] = line_items[i].pop("quantity")

                    if 'type' in line_items[i].keys():
                        line_items[i].pop("type")
                    if 'unit_of_measure' in line_items[i].keys():
                        line_items[i].pop("unit_of_measure")
                    if 'amount' in line_items[i].keys():
                        line_items[i].pop("amount")
                    if 'chart_of_account' in line_items[i].keys():
                        line_items[i].pop("chart_of_account")
                    if 'tax_type' in line_items[i].keys():
                        line_items[i].pop("tax_type")
                    if 'tax_id' in line_items[i].keys():
                        line_items[i].pop("tax_id")
                    if 'account_id' in line_items[i].keys():
                        line_items[i].pop("account_id")
                        
                if 'currency_code' in payload.keys():
                    payload.pop("currency_code")  
                if 'amount' in payload.keys():
                    payload.pop("amount")               

                # if line_items has data in it
                if bool(line_items):
                    data.update({
                        "line_items": line_items
                    })

            # If client type is Quickbooks Online
            elif super().get_client_type() == super().QBO:

                if 'customer_ledger_id' in payload.keys():
                    data.update({
                        "VendorRef": {
                            "value": payload.get("customer_ledger_id", 0)
                        }
                    })
                
                if "currency_code" in payload.keys():
                    data.update({
                        "CurrencyRef": {
                            "value": payload.get("currency_code", "KES")
                        }
                    })

                line_items = {}

                line_items = payload.get("line_items", [])

                for i in range(len(line_items)):
                    line_items[i]["DetailType"] = "ItemBasedExpenseLineDetail"
                    line_items[i]["ItemBasedExpenseLineDetail"] = {
                        "ItemRef": {
                            "value": line_items[i].pop("item_ledger_id")
                        },
                        "Qty": f'{line_items[i].get("quantity",0)}',
                        "UnitPrice": f'{line_items[i].get("price",0)}',
                    }
                    # line_items[i]["Qty"] = line_items[i].get("quantity",0)
                    # line_items[i]["UnitPrice"] = line_items[i].get("price",0)

                    if i==0:
                        if "amount" in line_items[i].keys():
                            line_items[i]["Amount"] = round(float(line_items[i].pop("amount")),2)
                        else:
                            line_items[i]["Amount"] = round(float(line_items[i].pop("quantity"))*float(line_items[i].pop("price")),2)

                    if "type" in line_items[i].keys():
                        line_items[i].pop("type")
                    if "item_id" in line_items[i].keys():
                        line_items[i].pop("item_id")
                    if "unit_of_measure" in line_items[i].keys():
                        line_items[i].pop("unit_of_measure")
                    if "account_id" in line_items[i].keys():
                        line_items[i].pop("account_id")
                    if "description" in line_items[i].keys():
                        line_items[i].pop("description")
                    if "tax_type" in line_items[i].keys():
                        line_items[i].pop("tax_type")
                    if "tax_id" in line_items[i].keys():
                        line_items[i].pop("tax_id")

                # if line_items has data in it
                if bool(line_items):
                    data.update({
                        "Line": line_items
                    })

            # If client type is MS_DYNAMICS
            elif super().get_client_type() == super().MS_DYNAMICS:
                # data.pop("type")
                data.update({
                    "vendor_number": f'{payload.get("vendor_id", "")}',
                    "vendor_invoice_no": f'{payload.get("vendor_invoice_no", "")}',
                    "due_date": f'{payload.get("due_date", "")}',
                })
                
                if "currency_code" in payload.keys():
                    payload.pop("currency_code")
                if "amount" in payload.keys():
                    payload.pop("amount")

                line_items = payload.get("line_items", [])

                for i in range(len(line_items)):
                    if "item_id" in line_items[i].keys():
                        line_items[i]["item_id"] = line_items[i].pop(
                            "item_id"
                        )
                    if "quantity" in line_items[i].keys():
                        line_items[i]["quantity"] = line_items[i].pop("quantity")
                    if "price" in line_items[i].keys():
                        line_items[i]["price"] = line_items[i].pop("price")
                    if "type" in line_items[i].keys():
                        line_items[i]["type"] = line_items[i].pop("type")
                    if "unit_of_measure" in line_items[i].keys():
                        line_items[i]["unit_of_measure"] = line_items[i].pop("unit_of_measure")

                    if "amount" in line_items[i].keys():
                        line_items[i].pop("amount")
                    if "chart_of_account" in line_items[i].keys():
                        line_items[i].pop("chart_of_account")

                # if line_items has data in it
                if bool(line_items):
                    data.update({"create_vendor_invoice_details": line_items})

            # If client type is XERO
            elif super().get_client_type() == super().XERO:

                invoices = {
                    "Type": "ACCPAY",
                    "Status": f'{additional_properties.get("Status", "DRAFT")}',
                    "Date": payload.get("invoice_date", datetime.now().strftime("%Y-%m-%d")),
                    "DueDate": payload.get("due_date", (datetime.now() + timedelta(days=1)).strftime("%Y-%m-%d")),
                }

                if "currency_code" in payload.keys():
                    invoices.update(
                        {"CurrencyCode": payload.get("currency_code", "KES")}
                    )

                if "customer_ledger_id" in payload.keys():
                    invoices.update(
                        {
                            "Contact": {
                                "ContactID": payload.get("customer_ledger_id", "")
                            }
                        }
                    )

                line_items = payload.get("line_items", [])

                for i in range(len(line_items)):
                    if "tax_type" in line_items[i].keys():
                        line_items[i]["TaxType"] = line_items[i].pop("tax_type")
                    elif "tax_type" not in line_items[i].keys():
                        line_items[i]["TaxType"] = "NONE"

                    if "price" in line_items[i].keys():
                        line_items[i]["UnitAmount"] = int(float(line_items[i].pop("price")))

                    if "quantity" in line_items[i].keys():
                        line_items[i]["Quantity"] = int(float(line_items[i].pop("quantity")))

                    line_items[i]["LineAmount"] = line_items[i].get("UnitAmount",0) * line_items[i].get("Quantity",0)

                    if "description" in line_items[i].keys():
                        line_items[i]["Description"] = line_items[i].pop("description")

                    if "account_id" in line_items[i].keys():
                        line_items[i]["AccountID"] = line_items[i].pop("account_id")

                    if "item_ledger_id" in line_items[i].keys():
                        line_items[i]["LineItemID"] = line_items[i].pop(
                            "item_ledger_id"
                        )

                    if "tax_id" in line_items[i].keys():
                        line_items[i].pop("tax_id")
                    if "type" in line_items[i].keys():
                        line_items[i].pop("type")
                    if "unit_of_measure" in line_items[i].keys():
                        line_items[i].pop("unit_of_measure")
                    

                if "reference" in payload.keys():
                    invoices.update({"Reference": payload.get("reference", "")})

                if "Status" in additional_properties.keys():
                    additional_properties.pop("Status")
                

                # if line_items has data in it
                if bool(line_items):
                    invoices.update({"LineItems": line_items})

                # if invoices has data in it
                if bool(invoices):
                    data.update({"Invoices": [invoices]})

            # If client type is ODOO
            elif super().get_client_type() == super().ODOO:
                data.update(
                    {
                        "move_type": "in_invoice",
                        "partner_id": int(payload.get("customer_ledger_id", 0)),
                        "name": payload.get("invoice_number", ""),
                        "payment_reference": payload.get("reference", ""),
                        "state": additional_properties.get("state", "draft"),
                        "currency_id": int(payload.get("currency_id", 0)),
                        "currency_code": payload.get("currency_code", ""),
                        "invoice_date": payload.get("invoice_date", datetime.now().strftime("%Y-%m-%d")),
                        "invoice_date_due": payload.get("due_date", (datetime.now() + timedelta(days=1)).strftime("%Y-%m-%d")),
                    }
                )

                line_items = payload.get("line_items", [])

                for i in range(len(line_items)):
                    if "item_ledger_id" in line_items[i].keys():
                        line_items[i]["product_id"] = int(line_items[i].pop("item_ledger_id"))
                    if "quantity" in line_items[i].keys():
                        line_items[i]["quantity"] = int(line_items[i].pop("quantity"))
                    if "price" in line_items[i].keys():
                        line_items[i]["price_unit"] = int(line_items[i].pop("price"))
                    if "tax_id" in line_items[i].keys():
                        line_items[i]["tax_ids"] = [int(line_items[i].pop("tax_id"))]

                    if "description" in line_items[i].keys():
                        line_items[i].pop("description")
                    if "account_id" in line_items[i].keys():
                        line_items[i].pop("account_id")
                    if "tax_type" in line_items[i].keys():
                        line_items[i].pop("tax_type")
                    if "type" in line_items[i].keys():
                        line_items[i].pop("type")
                    if "unit_of_measure" in line_items[i].keys():
                        line_items[i].pop("unit_of_measure")

                # if line_items has data in it
                if bool(line_items):
                    data.update({"invoice_line_ids": line_items})

            # If client type is ERP_NEXT
            elif super().get_client_type() == super().ERP_NEXT:
                data.update({
                    "supplier": f'{payload.get("customer_ledger_id", "")}',
                    "company": payload.get("company", ""),
                    "bill_no": payload.get("invoice_number", ""),
                    "bill_date": payload.get("invoice_date", ""),
                    "project": payload.get("project", ""),
                    "items": []
                })

                line_items = payload.get("line_items", [])

                for line_item in line_items:
                    data['items'].append({
                        "name": line_item.get("item_ledger_id", ""),
                        "item_code": line_item.get("item_ledger_id", ""),
                        "item_name": line_item.get("item_name", ""),
                        "qty": line_item.get("quantity", 0),
                        "description": line_item.get("description", ""),
                    })

            data.update(additional_properties)

            return data

        elif operation == super().READ:

            payload = super().response()

            data = payload

            # confirms if a single object was read from the database
            if isinstance(payload, dict):
                if 'resource' in payload.keys():
                    data = payload.get("resource", [])
                
            # confirms if a single object was read from the database
            if isinstance(data, dict):
                data = [data]
            
            # confirms if data is a list
            if isinstance(data, list):
                if len(data) > 0:
                    for i in range(len(data)):
                        if 'total_amount' in data[i].keys():
                            data[i]['amount'] = data[i].pop('total_amount')
                        if 'customer_id' in data[i].keys():
                            data[i]['vendor_id'] = data[i].pop('customer_id')
                        
            if 'resource' in payload.keys():
                payload["resource"] = data

            super().set_response(payload)

            return self


from erp_sync.Resources.resource import Resource
from erp_sync.Resources.client_types import ClientTypes
from erp_sync.Resources.customer import Customer
from erp_sync.Resources.company import Company
from erp_sync.Resources.items import Items
from erp_sync.Resources.taxes import Taxes
from erp_sync.Resources.chart_of_accounts import ChartOfAccounts
from erp_sync.Resources.currencies import Currencies
from erp_sync.Resources.bank_accounts import BankAccounts
from erp_sync.Resources.bank_transactions import BankTransactions
from erp_sync.Resources.bank_transfers import BankTransfers
from erp_sync.Resources.payments import Payments
from erp_sync.Resources.bill_payments import BillPayments
from erp_sync.Resources.bill_invoices import BillInvoices
from erp_sync.Resources.invoices import Invoices
from erp_sync.Resources.users import Users
from erp_sync.Resources.records import Records
from erp_sync.Resources.uploads import Uploads
from erp_sync.Resources.banks import Banks
from erp_sync.Resources.vendors import Vendors

class Client(Resource):

    _resources = {
        "ClientTypes": ClientTypes(),
        "Customer": Customer(),
        "Vendors": Vendors(),
        "Company": Company(),
        "Items": Items(),
        "Taxes": Taxes(),
        "ChartOfAccounts": ChartOfAccounts(),
        "Currencies": Currencies(),
        "Banks": Banks(),
        "BankAccounts": BankAccounts(),
        "BankTransactions": BankTransactions(),
        "BankTransfers": BankTransfers(),
        "Payments": Payments(),
        "BillPayments": BillPayments(),
        "BillInvoices": BillInvoices(),
        "Invoices": Invoices(),
        "Users": Users(),
        "Records": Records(),
        "Uploads": Uploads()
    }

    # use the nash object to confirm if the user accessing the client is logged in
    _nash = None

    urls = {}

    def __init__(self, nash, client_id=-1):
        self._nash = nash
        self._resources["Users"] = Users(self._nash)
        super().__init__("ClientsAPI", self._nash.get_headers(), self._nash.get_params())
        self.set_client_id(client_id)
        self._set_urls()

    def resource(self, resource_name):
        resource = self._resources[resource_name].set_client_id(super().get_client_id(
        )).set_company_id(super().get_company_id()).set_client_type(super().get_client_type())

        if resource_name == "Uploads":

            headers = self._nash.get_headers()

            if isinstance(headers, dict):
                if 'Content-Type' in headers.keys():
                    headers.pop("Content-Type")

            resource.set_headers(headers)
        else:
            resource.set_headers(self._nash.get_headers())

        return resource

    def get_resources(self):
        return list(self._resources.keys())

    def _set_urls(self):

        self.urls = {
            "new": f"/clients",
            "edit": f"/clients/{super().get_client_id()}",
            "read": f"/clients",
            "delete": f"/clients/{super().get_client_id()}",
        }

        super().set_urls(self.urls)

        return self

    def delete(self, payload=None, erp_id=None):

        self._set_urls()

        if erp_id is not None:
            self.urls["edit"] = f'{self.urls["edit"]}/{erp_id}'
            super().set_urls(self.urls)

        return super().delete(payload)

    def edit(self, payload=None, erp_id=None, method='PUT', endpoint=None):

        self._set_urls()

        if erp_id is not None:
            self.urls["edit"] = f'{self.urls["edit"]}/{erp_id}'
            super().set_urls(self.urls)

        return super().edit(payload)

    def read(self, erp_id=None, payload=None, method='GET', endpoint=None):

        self._set_urls()

        if erp_id is not None:
            self.urls["read"] = f'{self.urls["read"]}/{erp_id}'
            super().set_urls(self.urls)

        return super().read(payload, method, endpoint)

    def new(self, payload=None, method='POST', endpoint="/clients", add_client=True):
        # set data to pass to ruby url
        # set response here

        # Call the method exec to make the request to A.P.I. endpoint and return the data that will be returned in this method
        response = super().new(payload, method, endpoint).response()
        if response["status_code"] == 200:

            self.set_client_id(response["response_data"]["id"])

            self._set_urls()

            super().set_response(response)

        return self

    def get_callback_url(self, erp_id=None, payload=None, method='GET', endpoint='/authenticate/url'):

        if erp_id is None:

            erp_id = super().get_client_id()

        return super().read(payload, method, endpoint=f'{self.urls["new"]}/{erp_id}{endpoint}')

    def get_webhook_url(self, erp_id=None, payload=None, method='GET', endpoint='/hooks/url'):

        if erp_id is None:

            erp_id = super().get_client_id()

        return super().read(payload, method, endpoint=f'{self.urls["new"]}/{erp_id}{endpoint}')

    def authenticate(self, erp_id=None, payload=None, method='POST', endpoint='/authenticate'):

        if erp_id is None:

            erp_id = super().get_client_id()

        return super().read(payload, method, endpoint=f'{self.urls["new"]}/{erp_id}{endpoint}')

    def check_token(self, erp_id=None, payload=None, method='GET', endpoint='/authenticate/token'):

        if erp_id is None:

            erp_id = super().get_client_id()

        return super().read(payload, method, endpoint=f'{self.urls["new"]}/{erp_id}{endpoint}')

    def client_types(self, client_type_id=None):

        return self.resource("ClientTypes").read(client_type_id=client_type_id)

    def set_company_id(self, company_id=None):

        if company_id is not None:
            response = self.resource("Company").read(
                company_id=company_id).response()

            if bool(response):
                try:
                    super().set_client_id(response.get("client_id", None))
                    super().set_client_type(response.get("client_type_id", None))
                except Exception as e:
                    print(f"A ledger.nashglobal response error: {e} with response {response}")                
                self._set_urls()

        return super().set_company_id(company_id)



from erp_sync.Resources.customer import Customer
from erp_sync.Resources.company import Company
from erp_sync.APIs.api_format import API
import json

class Client(API, object):

    _resources = {
        "Customer": Customer(),
        "Company": Company()
    }

    # use the nash object to confirm if the user accessin the client is logged in
    _nash = None

    _response = {}

    _client_id = -1

    _company_id = 1

    def __init__(self,nash):
        self._nash = nash
        super().__init__("ClientsAPI", self._nash.get_headers(), self._nash.get_params())

    def resource(self,resource_name):
        # _resource = self._resources[resource_name].set_client_id(self.get_client_id())
        # _resource.set_headers(self._nash.get_headers())
        return self._resources[resource_name].set_client_id(self.get_client_id()).set_headers(self._nash.get_headers())
    
    def get_resources(self):
        return list(self._resources.keys())
    
    def get_client_id(self):
        return self._client_id
    
    def set_client_id(self, client_id):
        self._client_id = client_id
        return self
    
    def set_company_id(self,company_id):
        self._company_id = company_id
        return self
    
    def get_company_id(self):
        return self._company_id
    
    def authenticate(self,client_id = None,payload = None,method='POST',endpoint=""):
        if client_id is None:
            endpoint = f"/clients/{self.get_client_id()}/authenticate"
        else:
            endpoint = f"/clients/{client_id}/authenticate"
            self.set_client_id(client_id)

        
        # Call the method exec to make the request to A.P.I. endpoint and return the data that will be returned in this method
        self._response = self._exec(payload, method, endpoint)
        
        return self

    def new(self, payload = None, method='POST',endpoint="/clients"):
        # set data to pass to ruby url
        # set response here
        # set client id here
        # self.set_client_id()

        # Call the method exec to make the request to A.P.I. endpoint and return the data that will be returned in this method
        self._response = self._exec(payload, method, endpoint)
        if self._response["status_code"] == 200:

            self.set_client_id(self._response["response_data"]["id"])


        return self

    def read(self,data):
        # read data from ruby url
        # set response here
        return self
    
    def response(self):
        return self._response
    
    # This is the method that will be called execute an A.P.I. request.
    # Since most of the A.P.I. calls methods are similar, they are to be placed inside this method to avoid code duplication.
    # 
    # It will only accept parameters unique to each A.P.I. request. 
    def _exec(self, payload = None, method='POST', endpoint = ""):

        # Call the iPay A.P.I. url by passing the variables to the super class method responsible for making requests to A.P.I. endpoints
        # The super class method returns a response that is returned by this method
        return super().api_request(url=f"{super().get_base_url()}{endpoint}", payload=json.dumps(payload), method=method)
        



    
